#ifndef _ACCUMULATE_H_
#define _ACCUMULATE_H_

#include <numeric>
#include "../tbb_adapt.h"
#include "../jump_work.h"

namespace tbb_adapt {

template<class In, class T, class BinOp>
class AccumulateWork : public tbb_adapt::WorkAdapt
{
    In _first;
    In _last;
    T _local_res;
    BinOp _op;
    int _beg;
    int _end;
    int _pargrain;
    int _c;
    int _seqgrain;
    int _beg_local;
    int _end_local;

    public:

    AccumulateWork(In first, In last, T init, BinOp op, int beg, int c =1, int pargrain=2):
       WorkAdapt(), _first(first),  _last(last),  _local_res(init), _op(op),
         _beg(beg), _end(last - first), _pargrain(pargrain), _c(c)
    {
           if(c <=0) _seqgrain = (-c+1);
           else _seqgrain = _c * tbb_adapt::log_2(_end);
    }


    bool extract_nextseq()
    {
        
        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
           return true;
         } 
       return false;

    }


    bool extract_par( WorkAdapt*& sw)
    {
           if((_end - _beg) > _pargrain) {
             int i, j;
             int mid = _end - ((_end - _beg)/2);

             
             i = mid;
             j = _end;
             _end = mid;
             sw = new  AccumulateWork<In, T, BinOp>(_first, _first+j, T(0), _op, i, _c, _pargrain);
            return true;
           }
           else {
            sw = 0;
           return false;
          }
    }


    bool is_empty() {
        return (_beg >= _end);
    }

    void join(const WorkAdapt* jw)
    {
        const  AccumulateWork * src = dynamic_cast< const  AccumulateWork*> ( jw ); 
    
          T thief_res = src->get_result();
          _local_res +=thief_res ; 
    }

    void  localcompute()
    { 
      _local_res = std::accumulate(_first + _beg_local, _first + _end_local, _local_res, _op) ;
    }

    inline bool  MustJump() const { return true; }

    T get_result() const
    {
        return _local_res ;
    }

    void setjump(tbb_adapt::NextJump*& j) {
        
          int end = _end;
          _end = _beg;

          j = new NextJumpWork(_beg, end);   
                
   }

   void jump(const tbb_adapt::NextJump* j) {
        
        const  tbb_adapt::NextJumpWork * nextj = dynamic_cast< const  tbb_adapt::NextJumpWork*> ( j );
        _beg = nextj->get_beg();
        _end = nextj->get_end();
   }

};

template <typename InputIterator, typename T, typename BinOp>
inline T accumulate(InputIterator first, InputIterator last, T init,
      BinOp op, int seqgrain=100, int pargrain=2) 
{
     AccumulateWork<InputIterator, T, BinOp>* work;
     work = new  AccumulateWork<InputIterator, T, BinOp>(first, last, init, op, 0, seqgrain, pargrain);
     tbb_adapt::run_adapt(work);
     T res = work->get_result();
    return res ; 
}

template <typename InputIterator, typename T>
inline T accumulate(InputIterator first, InputIterator last, T init, int seqgrain=100, int pargrain=2)
{
   return tbb_adapt::accumulate(first, last, init, std::plus<T>(), seqgrain, pargrain);
}
}
#endif
