#ifndef _FIND_H_
#define _FIND_H_

#include <numeric>
#include "../tbb_adapt.h"
#include "../jump_work.h"


namespace tbb_adapt {

unsigned int find_if_macroloop_initsize = 300 ; //defaul value 

template<class InputIterator, class Predicate>
class FindIfWork : public tbb_adapt::WorkAdapt
{

    InputIterator _first;
    InputIterator _last;
    Predicate _pred;
    int _pos_found;
    int _beg;
    int _end;
    const int _init_beg;
    const int _init_end;
    int _seqgrain;
    int _pargrain;
    int _c;
    int _beg_local;
    int _end_local;
    int _thieft_pos_found;
    bool _element_found;   
 
    public:

     FindIfWork(InputIterator first, InputIterator last, Predicate pred, int beg, int c =1, int pargrain=2): 
       WorkAdapt(), _first(first),  _last(last), _pred(pred), _pos_found(last-first),
         _beg(beg), _init_beg(beg), _init_end(last - first), _pargrain(pargrain), _c(c)
    {
            _end = _init_end;
           if(c <=0) _seqgrain = 1;
           else _seqgrain = _c * tbb_adapt::log_2(_end);
           _thieft_pos_found = -1;
           _element_found = false;
    }


    bool extract_nextseq()
    {
        if(_element_found) {
             _beg = _end; //on s'arrête de continuer
            return false;
        }
    
        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
           return true;
         } 
       return false;

    }


    bool extract_par( WorkAdapt*& sw)
    {

           if((_end - _beg) > _pargrain) {
             int i, j;
             int mid = _end - ((_end - _beg)/2);

             
             i = mid;
             j = _end;
             _end = mid;
             sw = new  FindIfWork<InputIterator, Predicate>(_first, _first+j, _pred, i, _c, _pargrain);
            return true;
           }
           else {
            sw = 0;
           return false;
          }
    }

     bool  extract_next_macrostep()
     {
          static unsigned int alpha = find_if_macroloop_initsize ;
          bool extractmacro = ((_pos_found == _init_end) && (_beg  < _init_end));
          if(extractmacro){

            unsigned int completed_size = _beg - _init_beg ;
            int lg_sz = tbb_adapt::log_2(tbb_adapt::log_2( completed_size + 4 ) ) ;
            int size_macrostep= alpha*completed_size / lg_sz ;
            if (size_macrostep < find_if_macroloop_initsize)
                size_macrostep = find_if_macroloop_initsize ;
            _end = _beg + size_macrostep ;

            if(_end > _init_end) _end = _init_end;
             std::cout << "Macrostep:  _beg = " << _beg << ", _end=" << _end << std::endl;
          }
        return extractmacro;
      }


    void join(const WorkAdapt* jw)
    {
        const  FindIfWork<InputIterator, Predicate> *src = dynamic_cast< const  FindIfWork<InputIterator, 
               Predicate>*> ( jw ); 
    
          int thieft_pos_found = src->get_pos_found();
          int thieft_end = src->get_end();
          if(thieft_pos_found < thieft_end) _thieft_pos_found = thieft_pos_found;
         
    }

    bool is_empty() {
      return (_beg >= _end);
    }

    void  localcompute()
    { 
       InputIterator tmp = std::find_if(_first + _beg_local, _first + _end_local, _pred) ;

       if(tmp <  (_first + _end_local)) {
            _pos_found = tmp-_first;
            _element_found = true; //on arrete le calcul 
       }      
    }

    inline bool  MustJump() const { return true; }

    InputIterator get_result() const
    {
        return (_first+_pos_found) ;
    }

    int get_pos_found() const
    {
        return _pos_found;
    }

    int get_end() const 
    {
      return _end;
    }

    void setjump(tbb_adapt::NextJump*& j) {
        
          int end = _end;
          _end = _beg;
          j = new tbb_adapt::NextJumpWork(_beg, end);   
                
   }

   void jump(const tbb_adapt::NextJump* j) {
        
        const  tbb_adapt::NextJumpWork * nextj = dynamic_cast< const  tbb_adapt::NextJumpWork*> ( j );
        
        if((_pos_found==_init_end) && (_thieft_pos_found!=-1)) {
            std::cout << "trouve chez mon voleur...." << std::endl;
            _pos_found =  _thieft_pos_found;
            _beg = _end = nextj->get_end(); // on vide le travail restant
        }
        
        if(_pos_found == _init_end) {
           _beg = nextj->get_beg();
           _end = nextj->get_end();
        }
        else {
            _beg = _end = nextj->get_end();
        }

   }

};

template<class InputIterator, class Predicate>
inline InputIterator find_if(InputIterator first, InputIterator last, Predicate pred,
      int seqgrain=100, int pargrain=2) 
{
     FindIfWork<InputIterator, Predicate>* work = new  FindIfWork<InputIterator, Predicate>(first, last, pred, 0, 
         seqgrain, pargrain);
     tbb_adapt::run_adapt(work);
    return work->get_result(); ; 
}

}
#endif
