#ifndef _MERGE_H_
#define _MERGE_H_

#include <numeric>
#include "../tbb_adapt.h"
#include "../jump_work.h"


namespace tbb_adapt {

template<class InputIterator1, class InputIterator2, class OutputIterator, class Compare >
class MergeWork : public tbb_adapt::WorkAdapt
{
    InputIterator1 _first1;
    InputIterator1 _last1;
    InputIterator2 _first2;
    InputIterator2 _last2;
    OutputIterator _res;
    int _beg;
    int _beg2;
    Compare _cmp;
    int _end;
    int _end2;
    int _c;
    int _pargrain;
    int _sz;
    int _seqgrain;
   
    int _beg_local;
    int _end_local;
    int _beg_local2;
    int _end_local2;
    bool _signal_stop;
    bool _Iam_stopped;
    bool _thieft_stopped;
    pthread_mutex_t _my_mutex;
    bool _wait;
    int _thieft_beg1;
    int _thieft_beg2;

    public:

     MergeWork(InputIterator1 first1, InputIterator1 last1, InputIterator2 first2, InputIterator2 last2, 
         OutputIterator res, int beg, int beg2, Compare cmp, int c =1, 
              int pargrain=2, int sz=0): WorkAdapt(), _first1(first1),  _last1(last1), _first2(first2),
              _last2(last2), _res(res), _beg(beg), _beg2(beg2), _cmp(cmp), 
             _end(last1 - first1), _end2(last2-first2), _pargrain(pargrain), _c(c), _sz(sz)
    {
           if(c <=0) _seqgrain = (-c + 1);
           else {
                  if(_end > _end2) _seqgrain = _c*tbb_adapt::log_2(_end);
                  else _seqgrain = _c*tbb_adapt::log_2(_end2);
           }
           _beg_local = _beg;
           _end_local = _beg;

           _beg_local2 = beg2;
           _end_local2 = _beg2;
           _signal_stop = false;
           _Iam_stopped = false;
           _thieft_stopped = false;
           pthread_mutex_init(&_my_mutex, NULL);
           _wait = false;
    }


    bool extract_nextseq()
    {
            _wait = true;
        
           bool test_end1 = (_beg < _end) && (_beg_local >= _end_local);
           bool test_end2 = (_beg2 < _end2) && (_beg_local2 >= _end_local2);
           bool test3 = (_beg < _end) && (_beg_local < _end_local);
           bool test4 = (_beg2 < _end2) && (_beg_local2 < _end_local2);

          if(_signal_stop) { _Iam_stopped=true; }

          if( test_end1 && test_end2 )  {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
            
            _beg_local2 = _beg2; _beg2 +=_seqgrain;
            if( _beg2 > _end2)  _beg2 = _end2;
            _end_local2 = _beg2;
           
             return true;
           } 
           else if(test_end1) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
             return true;
           }
           else if(test_end2) {
            _beg_local2 = _beg2; _beg2 +=_seqgrain;
            if( _beg2 > _end2)  _beg2 = _end2;
            _end_local2 = _beg2;
             return true;
           }
           else if((_beg >= _end) && (_beg_local2 < _end_local2) && !_signal_stop ) {
             _beg2 = _end2;
            _end_local2 = _beg2;
            return true;
           }
           else if((_beg2 >= _end2) && (_beg_local < _end_local) && !_signal_stop ) {
            _beg = _end;
            _end_local = _beg;
            return true;
           }
           else if((test3 || test4) && !_signal_stop) {
                     return true;
           }
           else { _wait=false; return false;}
    }


    bool extract_par(WorkAdapt*& sw)
    {
        int size1 = (_end - _beg);
        int size2 =  (_end2-_beg2);
        if( size1 > _pargrain || size2 > _pargrain) {
          
          if(size1 > size2) {
            int i1, j1, i2, j2;
            int mid = _end - ((_end - _beg)/2);
            InputIterator2  split2 = std::lower_bound(_first2+_beg2, _first2+_end2, *(_first1+mid), _cmp);
            ptrdiff_t ptr_diff = (split2 - _first2);
            i2 = ptr_diff;
            if(i2<=_beg2) {
                   bool tmp;
                  pthread_mutex_lock(&_my_mutex);
                     tmp = _wait;
                  pthread_mutex_unlock(&_my_mutex);
                 static int cpt=0;
                 while(tmp) {
                   pthread_mutex_lock(&_my_mutex);
                     tmp = _wait;
                   pthread_mutex_unlock(&_my_mutex);
                   sched_yield(); 
                 }
                 _wait = false;
                 InputIterator2  split2 = std::lower_bound(_first2+_beg_local2, _first2+_beg2, *(_first1+mid), _cmp);
                 ptr_diff = (split2 - _first2);
                 i2 = ptr_diff;
                 _beg2 = _beg_local2;
                 _end_local2=_beg_local2;
            }
            i1 = mid;
            j1 = _end;
            _end = mid;
            j2 = _end2;
            _end2 = i2;

            int sz = (_end + _end2);
            sw = new MergeWork<InputIterator1, InputIterator2, OutputIterator, Compare>(_first1, _first1+j1,
                   _first2, _first2+j2, _res, i1, i2, _cmp, _c, _pargrain, sz);
          }
          else {
                int i1, j1, i2, j2;
                int mid = _end2 - ((_end2 - _beg2)/2);
                InputIterator1  split2 = std::upper_bound(_first1+_beg, _first1+_end, *(_first2+mid), _cmp);
                ptrdiff_t ptr_diff = (split2 - _first1);
                i1 = ptr_diff;
                if(i1<=_beg) {
                   bool tmp;
                  pthread_mutex_lock(&_my_mutex);
                     tmp = _wait;
                  pthread_mutex_unlock(&_my_mutex);
                 static int cpt = 0;
                 while(tmp) {
                   pthread_mutex_lock(&_my_mutex);
                     tmp = _wait;
                   pthread_mutex_unlock(&_my_mutex);
                   sched_yield();
                 }
                 _wait = false;
                 InputIterator1  split2 = std::upper_bound(_first1+_beg_local, _first1+_beg, *(_first2+mid), _cmp);
                 ptr_diff = (split2 - _first1);
                 i1 = ptr_diff;
                 _beg = _beg_local;
                 _end_local=_beg_local;
            }
            i2 = mid;
            j2 = _end2;
            _end2 = mid;
            j1 = _end;
            _end = i1;
            int sz = (_end + _end2);
            sw = new MergeWork<InputIterator1, InputIterator2, OutputIterator, Compare>(_first1, _first1+j1,
                  _first2, _first2+j2, _res, i1, i2, _cmp, _c, _pargrain, sz);
           }

         return true;
        }
        else {
            sw = 0;
         return false;
        }
    }

    bool is_empty()
    {
      return (_beg >= _end) && (_beg2 >= _end2);
    }

    void  localcompute()
    {
 
         InputIterator1 first1 = _first1+_beg_local;
         InputIterator1 last1 = _first1+_end_local;
         InputIterator2 first2 = _first2+_beg_local2;
         InputIterator2 last2 = _first2+_end_local2;
         OutputIterator res = _res + _sz;
         if(first1 >= last1) {
            int size2= (_end_local2-_beg_local2);
            //memcpy(res, first2, sizeof(T)*(size2));
            std::copy(first2, last2, res);
            _sz += (size2);
            _beg_local2 =  _end_local2 ;
          }
          else if(first2 >= last2) {
            int size1= (_end_local-_beg_local);
            //memcpy(res, first1, sizeof(T)*(size1));
            std::copy(first1, last1, res);
            _sz += size1;
           _beg_local =  _end_local ;
         }
         else {

             while((first1!=last1) && (first2!=last2)) {
                 if(_cmp(*first2, *first1))  *res++ = *first2++;
                 else *res++ = *first1++;
             }
            
             _beg_local = (first1-_first1);
             _beg_local2 = (first2-_first2);
             _sz = (_beg_local + _beg_local2);
        }
 
         pthread_mutex_lock(&_my_mutex);
            _wait = false;
         pthread_mutex_unlock(&_my_mutex);
     }


    void join(const WorkAdapt* jw)
    {
        const  MergeWork * src = dynamic_cast< const  MergeWork*> ( jw );

         int thieft_sz = src->get_sz();
         _thieft_beg1 = src->get_beg();
         _thieft_beg2 = src->get_beg2();
         _thieft_stopped = src->get_stop();
         _sz =  thieft_sz;
    }



    int get_sz() const { return _sz; }
    bool get_stop() const { return _Iam_stopped; }
  
    int get_beg() const { return _beg_local;} 
    int get_beg2() const { return _beg_local2;}

   bool  MainJump() const { return false; }
   // inline bool  MustJump() const { return true; }

    void setjump(tbb_adapt::NextJump*& j) {
          int beg, beg2;        
         
           beg = _beg;
          _beg = _end;
          
           beg2 = _beg2;
          _beg2 = _end2;
         
           j = new tbb_adapt::NextJumpWork2d(beg, _end, beg2, _end2);  
           _signal_stop = true; 
                
   }

   void jump(const tbb_adapt::NextJump* j) {
        
        const  tbb_adapt::NextJumpWork2d * nextj = dynamic_cast< const  tbb_adapt::NextJumpWork2d*> ( j );
        
       if(!_thieft_stopped) {
              _beg = nextj->get_beg();
              _beg2 = nextj->get_beg2();
              _end = nextj->get_end();
              _end2 = nextj->get_end2();
        }
        else {
        _beg =  _thieft_beg1;
        _beg2 = _thieft_beg2; 
        _end = nextj->get_end();
        _end2 = nextj->get_end2();
        }
        
   }


};


template <typename InputIterator1, typename InputIterator2, typename OutputIterator, typename Compare>
inline OutputIterator merge(InputIterator1 first1, InputIterator1 last1, InputIterator2 first2, 
      InputIterator2 last2, OutputIterator output, Compare cmp, int seqgrain=100, int pargrain=2)
{
     MergeWork<InputIterator1, InputIterator2, OutputIterator, Compare>* work ;
     work = new  MergeWork<InputIterator1, InputIterator2, OutputIterator, Compare>(first1, last1, first2, last2, 
          output, 0, 0, cmp, seqgrain, pargrain);
    tbb_adapt::run_adapt(work);

return output + ((last1-first1) + (last2-first2));
}


template <typename InputIterator1, typename InputIterator2, typename OutputIterator>
inline OutputIterator merge(InputIterator1 first1, InputIterator1 last1, InputIterator2 first2, 
    InputIterator2 last2, OutputIterator output, int seqgrain=100, int pargrain=2) 
{
  typedef typename std::iterator_traits<OutputIterator>::value_type val_t;  
  return tbb_adapt::merge( first1, last1, first2,  last2, output, std::less<val_t>(), seqgrain, pargrain);    
}

}
#endif
