
#include <iostream>
#include <numa.h>
#include "parallel_scan_adapt.h"

typedef double val_t ;

#include "tbb/task_scheduler_init.h"
#include "tbb/tick_count.h"
#include <tbb/blocked_range.h>

template<class T>
bool verif (T first, T last, T res) {
     int i = 0;
     int cpt = 0;
     bool is_ok = true;
     if(((*first-*res) > 1e-6) || ((*res-*first) > 1e-6) ) {
      std::cerr << "Erreur: res[" << i << "]=" << *res << "!=" << *first << std::endl;
      is_ok=false;
     }
     while(++first!=last) {
       ++res;
       i++;
       if(((*first-*res) > 1e-6) || ((*res-*first) > 1e-6)) {
        cpt++;
        if(cpt < 10)  std::cerr << "Erreur: res[" << i << "]=" << *res << "!=" << *first << std::endl;
        is_ok=false;
      }
     }
 return is_ok;
}

template <class T, class op=std::plus<T> >
struct my_op {
  volatile double d;
    T operator()(T a, T b){
        /*for (int i=0; i<1000000; ++i)
           d = sin(1000);*/  
     return  op()(a, b);
    }
};


template<typename T>
class Body {

  T* const y;
  const T* const x;
  T local_res;
  T main_res;

public:

  Body(T y_[], const T x_[]) : y(y_), x(x_), local_res(0), main_res(0) {}

  template<typename Tag>
  void operator()(const tbb::blocked_range<int>& r, Tag) {

     T tmp = local_res;
    
    if(Tag::is_final_scan()) {
      //std::cout << "Scan Computation: [" << r.begin() << " " << r.end() << "[" << std::endl;
      for(int i=r.begin(); i < r.end(); ++i) {
         tmp =  my_op<T>()(tmp,  x[i]);
         y[i] = tmp;
      }
     local_res = tmp;
    }
    else {
        //std::cout << "Tranform Computation: [" << r.begin() << " " << r.end() << "[" << std::endl;
        for(int i=r.begin(); i < r.end(); ++i) {
          y[i] = my_op<T>()(main_res, y[i]);
        }
    } 

  }

   Body(Body& b, tbb::split) : y(b.y), x(b.x), local_res(0), main_res(b.main_res) { }

   T get_local_res() const { 
     return local_res; 
   }
  
   void join(const Body<T> & a) {
     local_res = my_op<T>()(local_res, a.local_res);
   }

   void get_main_result(const Body<T> & a) {
       main_res = a.local_res;
     std::cout << "main_res = " << main_res << std::endl;
   }

   void assign(int i) {
      y[i] = local_res;
   }
};

template<typename T>
T DoParallelPrefix( T* first, T* last, T* result, size_t c=100) {
  const int n = last - first;
  Body<T> body(result, first);
  parallel_scan_adapt( tbb::blocked_range<int>(0,n), body, c);
return body.get_local_res();
}


int main(int argc, char ** argv) {
      

       int size = 100000000;
       int v , help;
       size_t c = 100 ;//facteur par defaut
       int nthread = 1;
       
       char * specifiers [] = { (char*)"-n", (char*)"-c", (char*)"-m", (char*)"-np", (char*)"-v", (char*)"-h" ,0};
       int opt_types[] = {INTARG, INTARG, INTARG, INTARG, BOOLARG, BOOLARG, 0};
       get_options(argc, argv, specifiers, opt_types, &size, &c, &macroloop_initsize, &nthread, &v, &help, 0);

       val_t *data = (val_t*) numa_alloc_interleaved(size*sizeof(val_t));
       val_t *output = (val_t*) numa_alloc_interleaved(size*sizeof(val_t));
       val_t *output2 = (val_t*) numa_alloc_interleaved(size*sizeof(val_t));
       //std::cout << "size =" << size << std::endl;
       for(int i =0; i < size; i++) data[i] = i+1;
       
       tbb::task_scheduler_init init(nthread);
 
       tbb::tick_count t0, t1;
       //std::cout << "***********Start of computation**********" << std::endl;
       t0 = tbb::tick_count::now();
        val_t res = DoParallelPrefix(data, data+size, output, c) ;
       t1 = tbb::tick_count::now();
    
       std::cout << "-----------------------------------" << std::endl;
       printf (": time = %.6f s\n", (t1-t0).seconds());
       std::cout << "res = " << res << std::endl;
       std::cout << "-----------------------------------" << std::endl;
     

       if(v) { 
       //std::cout << "-------------------Debut verification---------------" << std::endl;
         val_t* output2 = new val_t[size];
         std::partial_sum(data, data+size, output2); //STL
         if(verif(output2, output2+size, output))
              std::cout << "Le calcul est correct : OK! " << std::endl; 
         else std::cout << "Le calcul n'est pas correct : KO! " << std::endl;
        //for(int i =0; i < size; i++) cout << "output[" << i << "]=" << output[i] << endl;
        delete [] output2;
        //std::cout << "-------------------Fin verification-----------" << std::endl;
       }
 
       numa_free(data, size*sizeof(val_t));    
       numa_free(output, size*sizeof(val_t));
  return 0;
};

