
#include <iostream>
#include <numa.h>
#include <numaif.h>
#include <sys/mman.h>
#include <vector> //Pour faire des tests
#include "partial_sum.h"
#include "../accumulate/accumulate.h"
#include "../transform/transform.h"
#include "../util/getoptions.h"

#include "tbb/task_scheduler_init.h"
#include "tbb/tick_count.h"

//#define NUMA_CTL
#define VECTOR

typedef double val_t ;

template<class T>
struct my_fct {
    my_fct(){ };
    T operator()(T a){
         return  2*a;
    }
};


template<class In, class Out>
bool verif (In first, In last, Out res) {
     int i = 0;
     int cpt = 0;
     bool is_ok = true;
     if(((*first-*res) > 10e-3) || ((*res-*first) > 10e-3)) {
      std::cerr << "Erreur: res[" << i << "]=" << *res << "!=" << *first << std::endl;
      is_ok=false;
     }
     while(++first!=last) {
       ++res;
       i++;
       if(((*first-*res) > 10e-3) || ((*res-*first) > 10e-3)) {
        cpt++;
        if(cpt < 10)  std::cerr << "Erreur: res[" << i << "]=" << *res << "!=" << *first << std::endl;
        is_ok=false;
      }
     }
 return is_ok;
}


int main(int argc, char ** argv) {
      
	    
       int size = 100000000;
       int v , s, help;
       int c = 100 ;//facteur par defaut
       int nthread = 1;
       
       char * specifiers [] = { (char*)"-n", (char*)"-c", (char*)"-m", (char*)"-np", (char*)"-s", (char*)"-v", 
        (char*)"-h" ,0};
       int opt_types[] = {INTARG, INTARG, INTARG, INTARG, BOOLARG, BOOLARG, BOOLARG, 0};
       get_options(argc, argv, specifiers, opt_types, &size, &c, &tbb_adapt::partial_sum_macroloop_initsize, 
           &nthread, &s, &v, &help, 0);

       #ifdef NUMA_CTL
       std::cout << "Computation pointer with NUMA_CTL, size = " << std::endl;
       unsigned long numa_node = 255;
       val_t *data = (val_t *)mmap(0, (size)*sizeof(val_t),PROT_READ|PROT_WRITE,
                                 MAP_PRIVATE|MAP_ANONYMOUS,0,0);
       if (mbind(data,size*sizeof(val_t),MPOL_INTERLEAVE,&numa_node,64,0) != 0)
       {
          perror("mbind");
       }

       val_t *output = (val_t *)mmap(0, (size)*sizeof(val_t),PROT_READ|PROT_WRITE,
                                 MAP_PRIVATE|MAP_ANONYMOUS,0,0);
       if (mbind(output, size*sizeof(val_t),MPOL_INTERLEAVE,&numa_node,64,0) != 0)
       {
          //perror("mbind");
       }
      #else 
        #ifdef VECTOR
        std::cout << "Computation with vector, size = " << size << std::endl; 
        std::vector<val_t> data(size);
        std::vector<val_t> output(size);
        #else
        std::cout << "Computation pointer without NUMA_CTL, size = " << size << std::endl;
        val_t *data =  new val_t[size];
        val_t *output = new val_t[size];
        #endif
      #endif

      for(int i =0; i < size; i++) {
          data[i] = i+1;
          output[i] = 0;
       }
       
       tbb::task_scheduler_init init(nthread);
 
       tbb::tick_count t0, t1;
       std::cout << "***********Start of computation**********" << std::endl;
       val_t res;
       t0 = tbb::tick_count::now();
       if(!s) {
        #ifndef VECTOR 
        res = *(tbb_adapt::partial_sum(data, data+size, output/*, c*/)-1) ;
        #else
        res = *(tbb_adapt::partial_sum(data.begin(), data.end(), output.begin()/*, c*/)-1) ;
        #endif
       } else {
         #ifndef VECTOR
        res = *(tbb_adapt::partial_sum(data, data+size, output/*, c*/)-1) ;
        #else
        res = *(std::partial_sum(data.begin(), data.end(), output.begin())-1) ;
        #endif
       }
       t1 = tbb::tick_count::now();
   
        
       std::cout << "-----------------------------------" << std::endl;
       printf (": time = %.6f s\n", (t1-t0).seconds());
       //printf (": time = %.0f ns\n", (1e+9)*(t1-t0).seconds());
       std::cout << ": res = " << res  << std::endl;
       std::cout << "-----------------------------------" << std::endl;
       

       if(v) { 
        std::cout << "-------------------Debut verification---------------" << std::endl;
         val_t* output2 = new val_t[size];
         #ifndef VECTOR
         std::partial_sum(data, data+size, output2); //STL
         if(verif(output2, output2+size, output))
         #else
         std::partial_sum(data.begin(), data.end(), output2); //STL
         if(verif(output2, output2+size, output.begin()))
         #endif
         std::cout << "Le calcul est correct : OK! " << std::endl; 
         else std::cout << "Le calcul n'est pas correct KO! " << std::endl;
        //for(int i =0; i < size; i++) cout << "output[" << i << "]=" << output[i] << endl;
        delete [] output2;
        std::cout << "-------------------Fin verification-----------" << std::endl;
       }
 
       #ifdef NUMA_CTL
       numa_free(data, size*sizeof(val_t));    
       numa_free(output, size*sizeof(val_t));
       #else
        #ifndef VECTOR
           delete [] data;
           delete [] output;
        #endif
       #endif
  return 0;
};

