#ifndef _PARTIAL_SUM_H_
#define _PARTIAL_SUM_H_

#include <numeric>
#include <functional>
#include "../tbb_adapt.h"
#include "../jump_work.h"

namespace tbb_adapt {

unsigned int partial_sum_macroloop_initsize = 300 ; //defaul value


//The Unary operation using by tranform algorithm
template <class T, class BinOp>
struct UnaryOp {
    T _val;
    BinOp _op;
    UnaryOp(T val, BinOp op):_val(val), _op(op) { };
    T operator()(T a){
         return  _op(_val, a);
    }
};

//The finalization Work
template<class T, class BinOp>
class Partial_Sum_FinalizeWork : public tbb_adapt::FinalizeWork {

    T _first;
    T _last;
    BinOp _op;
    int _beg;
    int _end;
    typedef typename std::iterator_traits<T>::value_type val_t;
    val_t _val;
    int _beg_local ;
    int _end_local;
    int _c;
    int _pargrain;
    int _seqgrain;

    public :
     Partial_Sum_FinalizeWork(T first, T last, BinOp op, int beg, val_t val=0, int c=100, int pargrain=2): 
       FinalizeWork(), _first(first),  _last(last), _op(op), _beg(beg), _end(last-first), _val(val), 
       _c(c), _pargrain(pargrain) 
     { 
        if(c <=0) _seqgrain = (-c+1);
        else _seqgrain = _c * tbb_adapt::log_2(_end);
     }
 

     bool extract_nextseq()
    {

        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
           return true;
         }
       return false;

    }


    bool extract_par( FinalizeWork*& sw)
    {
           if((_end - _beg) > _pargrain) {
             int i, j;
             int mid = _end - ((_end - _beg)/2);


             i = mid;
             j = _end;
             _end = mid;
             sw = new  Partial_Sum_FinalizeWork<T, BinOp>( _first, _first+j, _op, i, _val, _c, _pargrain);
            return true;
           }
           else {
            sw = 0;
            return false;
          }
     }

 
     void localcompute()
     {
       std::transform(_first + _beg_local, _first + _end_local, _first + _beg_local, UnaryOp<val_t, BinOp>(_val, _op)) ;
     }

};

/*Nano-loop Sequential computation with given op*/
template<class In, class Out, class BinOp>
Out partial_sum_local_compute (In first, In last, Out res, BinOp op) {
    typedef typename std::iterator_traits<Out>::value_type val_t;
    val_t val =  *res ;
    while(++first!=last){
        val=op(val, *first);
        *++res=val;
    }
    return res;
};

/*Nano-loop Sequential computation with defaul op*/
template<class In, class Out>
Out partial_sum_local_compute (In first, In last, Out res) {
  typedef typename std::iterator_traits<Out>::value_type val_t;
  return partial_sum_local_compute(first, last, res, std::plus<val_t>());
};

 
template<class In, class Out, class BinOp>
class PrefixWork : public tbb_adapt::WorkAdapt
{
    In _first;
    In _last;
    Out _res;
    BinOp _op;
    typedef typename std::iterator_traits<Out>::value_type val_t;
    val_t _local_res;
    int _beg;
    int _end;
    const int _init_beg;
    const int _init_end;
    int _seqgrain;
    int _pargrain;
    int _c;
    int _beg_local;
    int _end_local;
    val_t _main_res;
    val_t _thieft_res;
    public:

     PrefixWork(In first, In last, Out res, BinOp op, int beg, int c =100, int pargrain=2): 
       WorkAdapt(), _first(first),  _last(last), _res(res), _op(op), _local_res(val_t(0)),
         _beg(beg), _init_beg(beg), _init_end(last - first), _pargrain(pargrain), _c(c)
    {
           _end = _init_end;
           if(c <=0) _seqgrain = (-c+1);
           else _seqgrain = _c * tbb_adapt::log_2(_end);
    }


    bool extract_nextseq()
    {
        
        if( _beg < _end ) {
            _beg_local = _beg; _beg +=_seqgrain;
            if( _beg > _end)  _beg = _end;
            _end_local = _beg;
           return true;
         } 
       return false;

    }


    bool extract_par( WorkAdapt*& sw)
    {
           if((_end - _beg) > _pargrain) {
             int i, j;
             int mid = _end - ((_end - _beg)/2);
 
             i = mid;
             j = _end;
             _end = mid;
             sw = new  PrefixWork<In, Out, BinOp>(_first, _first+j, _res, _op, i, _c, _pargrain);
            return true;
           }
           else {
            sw = new  PrefixWork<In, Out, BinOp>(_first, _first, _res, _op,0);
            return false;
          }
    }

     bool  extract_next_macrostep()
     {
          static unsigned int alpha = partial_sum_macroloop_initsize ;
          bool extractmacro = (_beg  < _init_end);
          if(extractmacro){

            unsigned int completed_size = _beg - _init_beg ;
            int lg_sz = tbb_adapt::log_2(tbb_adapt::log_2( completed_size + 4 ) ) ;
            int size_macrostep= alpha*completed_size / lg_sz ;
            if (size_macrostep < partial_sum_macroloop_initsize)
                size_macrostep = partial_sum_macroloop_initsize ;
            _end = _beg + size_macrostep ;

            if(_end > _init_end) _end = _init_end;
          }
        return extractmacro;
      }

    bool is_empty() {
          return (_beg >= _end);
    }

    void join(const WorkAdapt* jw)
    {
        const  PrefixWork * src = dynamic_cast< const  PrefixWork*> ( jw ); 
    
          val_t thief_res = src->get_result();
          _local_res = _op(_local_res, thief_res) ;
          int thieft_end = src->get_end();
           *(_res+(thieft_end-1))   = _local_res;
    }

   void  get_main_result(const WorkAdapt* jw ) 
   {
        const  PrefixWork * src = dynamic_cast< const  PrefixWork*> ( jw );
        _main_res = src->get_result();
   }

    void  localcompute()
    { 
       val_t val = ( (_local_res==val_t(0))? *(_first+_beg_local) : _op(_local_res, *(_first+_beg_local)));
      *(_res+_beg_local)   = val ; 
       _local_res = *(partial_sum_local_compute(_first + _beg_local, _first + _end_local, _res+_beg_local, _op)) ;
    }

    inline bool is_finalize_work() const { return true; }

    void get_finalize_work(tbb_adapt::FinalizeWork*& fw)
    { 
        fw = new Partial_Sum_FinalizeWork<Out, BinOp>(_res, _res+(_beg-1), _op, _init_beg, _main_res, _c, _pargrain);
    }

    val_t get_result() const
    {
        return _local_res ;
    }

    int get_end() const 
    {
      return _end;
    }

    void setjump(tbb_adapt::NextJump*& j) {
        
          int end = _end;
          _end = _beg;

          j = new tbb_adapt::NextJumpWork(_beg, end);   
                
   }

   void jump(const tbb_adapt::NextJump* j) {
        
        const  tbb_adapt::NextJumpWork * nextj = dynamic_cast< const  NextJumpWork*> ( j );
        _beg = nextj->get_beg();
        _end = nextj->get_end();
   }

};

////////////////////Adaptive parallel partial_sum/////////////////////////////

////////////////////////////:partial_sum with seqgrain type equal to int

template <typename InputIterator,  typename OutputIterator, typename BinOp>
inline OutputIterator partial_sum(InputIterator first, InputIterator last, OutputIterator res, 
    BinOp op, int seqgrain=100, size_t pargrain=2) 
{
    if(first==last) return res;
    typedef typename std::iterator_traits<OutputIterator>::value_type val_t;
    PrefixWork<InputIterator, OutputIterator, BinOp>* work;
    work = new  PrefixWork<InputIterator, OutputIterator, BinOp>(first, last, res, op, 0, seqgrain, pargrain);
    tbb_adapt::run_adapt(work);
    
    return res + (last-first) ; 
}

template <typename InputIterator,  typename OutputIterator>
inline
OutputIterator partial_sum(InputIterator first, InputIterator last, OutputIterator res, int seqgrain=100, size_t pargrain=2)
{
 typedef typename std::iterator_traits<OutputIterator>::value_type val_t;
 return tbb_adapt::partial_sum(first, last, res, std::plus<val_t>(), seqgrain, pargrain);
}

//////////////////////////// partial_sum with seqgrain type equal to size_t
template <typename InputIterator,  typename OutputIterator, typename BinOp>
inline OutputIterator partial_sum(InputIterator first, InputIterator last, OutputIterator res,
    BinOp op, size_t seqgrain, size_t pargrain=2)
{
    if(first==last) return res;
    typedef typename std::iterator_traits<OutputIterator>::value_type val_t;
    PrefixWork<InputIterator, OutputIterator, BinOp>* work;
    work = new  PrefixWork<InputIterator, OutputIterator, BinOp>(first, last, res, op, 0, seqgrain, pargrain);
    tbb_adapt::run_adapt(work);

    return res + (last-first) ;
}

template <typename InputIterator,  typename OutputIterator>
inline 
OutputIterator partial_sum(InputIterator first, InputIterator last, OutputIterator res, size_t seqgrain, size_t pargrain=2)
{
 typedef typename std::iterator_traits<OutputIterator>::value_type val_t;
 return tbb_adapt::partial_sum(first, last, res, std::plus<val_t>(), seqgrain, pargrain);
}

}//end namespace tbb_adapt


#endif
