#ifndef _PARTITION_H_
#define _PARTITION_H_

/*Author : D Traore
 *Date : Fevrier 2009
*/


#include <numeric>
#include "../tbb_adapt.h"
#include "../jump_work.h"


namespace tbb_adapt {


///////////////////////////////Parallel adaptive partition with adaptive grain size//////////////////

template<class RandomAccessIterator, class Pred>
class PartitionWork : public tbb_adapt::WorkAdapt
{
     RandomAccessIterator _first;
     RandomAccessIterator _last;
     int _beg;
     int _beg2;
     const int _init_beg;
     const int _init_beg2;
     int _end;
     int _end2;
     Pred _p;
     int _init_end;;
     int _c;
     int _pargrain;
     int _sz;
     int _seqgrain;
   
    int _beg_local;
    int _end_local;
    int _beg_local2;
    int _end_local2;
    bool _signal_stop;
    bool _Iam_stopped;
    bool _thieft_stopped;
    pthread_mutex_t _my_mutex;
    bool _wait;
    int _thieft_beg1;
    int _thieft_beg2;
    std::list< const  tbb_adapt::NextJump* > B_F_left;
    std::list< const  tbb_adapt::NextJump* > B_F_right;

    public:

     PartitionWork(RandomAccessIterator first, RandomAccessIterator last, int beg, int beg2, int end, 
       Pred p, int c =1, int pargrain=2): WorkAdapt(), _first(first),  _last(last), _beg(beg), _beg2(beg2),
         _init_beg(_beg), _init_beg2(_beg2), _end(end), _end2(last - first), _p(p),_init_end(end), 
         _pargrain(pargrain), _c(c)
    {
           
           if(c <=0) _seqgrain = 1;
           else _seqgrain = _c*tbb_adapt::log_2(_end);

           _beg_local = _beg;
           _end_local = _beg;

           _beg_local2 = _end2;
           _end_local2 = _end2;
           _signal_stop = false;
           _Iam_stopped = false;
           _thieft_stopped = false;
           pthread_mutex_init(&_my_mutex, NULL);
           _wait = false;
    }


    bool extract_nextseq()
    {
            _wait = true;
        
          if( (_beg < _end) && (_beg2 < _end2) )  {
          
             if(_beg_local >= _end_local){
              _beg_local = _beg; _beg +=_seqgrain;
               if( _beg > _end)  _beg = _end;
              _end_local = _beg;
             }

             if(_beg_local2 >= _end_local2){
              _end_local2 = _end2; _end2 -=_seqgrain;
              if( _end2 < _beg2)  _end2 = _beg2;
             _beg_local2 = _end2;
             }
             return true;
          }
          else if((_beg < _end) && (_beg_local >= _end_local)){
            
               if(_beg_local2 < _end_local2){
                _beg_local = _beg; _beg +=_seqgrain;
                 if( _beg > _end)  _beg = _end;
                 _end_local = _beg; 
                 return true;
               }
               else {
                   _beg = _end;
                   _end_local = _beg;
  
               _wait=false;
               return false;  
              }
          }
           else if((_beg2 < _end2) && (_beg_local2 >= _end_local2)){
               if(_beg_local < _end_local){
                  _end_local2 = _end2; _end2 -=_seqgrain;
                  if( _end2 < _beg2)  _end2 = _beg2;
                 _beg_local2 = _end2;
                 return true;
               }
               else {
                     _end2 = _beg2;
                     _beg_local2 = _end2;
                   _wait=false; return false;
               }  
           } 
            else if((_beg >= _end) && (_beg_local >= _end_local)){
                     _end2 = _beg2;
                     _beg_local2 = _end2;
                   _wait=false; return false;
            } 
            else if((_beg2 >= _end2) && (_beg_local2 >= _end_local2)){
                    _beg = _end;
                    _end_local = _beg;
                   _wait=false; return false;
            }
           else {
               _wait=false;
            return false;
          }
    }


    bool extract_par(WorkAdapt*& sw)
    {
        int size1 = (_end - _beg);
        int size2 =  (_end2-_beg2);
        if( (size1 > _pargrain) && (size2 > _pargrain)) {
             int i1, j1;
             int i2, j2;
             int mid1 = _end - ((_end - _beg)/2);
             int mid2 = _end2 - ((_end2 - _beg2)/2);
             i1 = mid1;
             j1 = _end;
             _end = mid1;

             i2 = _beg2;
             j2 = mid2;
             _beg2 = mid2;

           #ifdef PRINT
            cout << "1:work : (i1, j1)=" << "("<< _beg << ", " << _end << ")" << std::endl;
            cout << "1:stolen work : (i1, j1)=" << "("<< i1 << ", " << j1 << ")" << std::endl;
            cout << "1:work : (i2, j2)=" << "("<< _beg2 << ", " << _end2 << ")" << std::endl;
            cout << "1:stolen work : (i2, j2)=" << "("<< i2 << ", " << j2 << ")" << std::endl;
           #endif
            
            sw = new PartitionWork<RandomAccessIterator, Pred>(_first, _first+j2, i1, i2, j1, _p, _c, _pargrain);

         return true;
        }
        else {
           sw = 0;
           return false;
        }
    }

    /*void join(const WorkAdapt* jw)
    {
    }*/

    bool is_empty() {
      return (_beg >= _end) && (_beg2 >= _end2);
    }

    void  localcompute()
    {

         RandomAccessIterator first1 = _first+_beg_local;
         RandomAccessIterator last1 = _first+_end_local;
         RandomAccessIterator first2 = _first+_beg_local2-1;
         RandomAccessIterator last2 = _first+_end_local2-1;
         
         do {
              while((first1!=last1) && _p(*first1)) first1++;
              while((last2!=first2) && !_p(*last2)) last2--;
              if((first1==last1) || (last2==first2)) break;
              std::swap(*first1, *last2);
              first1++ ; *last2--;
            } while((first1!=last1)&&(last2!=first2));
 
             _beg_local = (first1-_first);
             _end_local2 = (last2-_first)+1;
       
   
         pthread_mutex_lock(&_my_mutex);
            _wait = false;
         pthread_mutex_unlock(&_my_mutex);
     }



   RandomAccessIterator get_split() const { 

        if(_beg_local2 >= _end_local2) {
              return std::partition(_first+_beg_local, _first+_end_local, _p);
         }
         else {
             return  std::partition(_first+_beg_local2, _first+_end_local2, _p);
        }
  }   

   //bool  MainJump() const { return false; }
   //inline bool  MustJump() const { return true; }

    void setjump(tbb_adapt::NextJump*& j) {
        
         //attente
           bool tmp;
           pthread_mutex_lock(&_my_mutex);
               tmp = _wait;
           pthread_mutex_unlock(&_my_mutex);
           while(tmp) {
             pthread_mutex_lock(&_my_mutex);
                tmp = _wait;
             pthread_mutex_unlock(&_my_mutex);
             sched_yield();
           }
           _wait = false;
         
           j = new tbb_adapt::NextJumpWork2d(_beg_local, _end, _beg2, _end_local2);

          //On vide tout le travail restant
           _beg = _end;
           _end2 = _beg2;
           _beg_local = _end_local  =_beg;
           _end_local2 = _beg_local2=_end2;

    }

   void jump(const tbb_adapt::NextJump* j) {
        
        const  tbb_adapt::NextJumpWork2d * nextj = dynamic_cast< const  tbb_adapt::NextJumpWork2d*> ( j );

         #ifdef PRINT2
         static int nb_sauts=0;
         ++nb_sauts; 
         static int nb_treatment_left = 0;
         static int nb_treatment_right = 0;
         static int nb_swap_total = 0;
        #endif 

         if(_beg_local < _end_local) {
           
             int i1 = nextj->get_beg() ;
             int i2 = nextj->get_end() ;
             int treated_elts_size = i1 - _end_local;
             int untreated_elts_size = _end_local - _beg_local;

                    
             int k= nextj->get_beg2();//AJOUT
             bool end_treatment = ((k==_init_end) || (i2==_init_end));//AJOUT
           
             if(((i1 < i2) && (treated_elts_size > 0)) || end_treatment) { //AJOUT

              #ifdef PRINT2
              ++nb_treatment_left; //AJOUT
              #endif

              int n_swap; int i;
              int tmp;
              
              if(untreated_elts_size < treated_elts_size)  {
                        n_swap = untreated_elts_size; i=tmp=i1-untreated_elts_size;
              }
              else { n_swap = treated_elts_size; i=_end_local; tmp=_beg_local+treated_elts_size;}

              std::swap_ranges(_first+_beg_local, _first+_beg_local+n_swap, _first+i);
              _beg = tmp; _end = i2; 
              _beg_local = _end_local = _beg;
             
             #ifdef PRINT2
             nb_swap_total +=n_swap; //AJOUT
             #endif
             } else {
                if(treated_elts_size==0) {
                   _beg = _beg_local; _end = i2;
                   _beg_local = _end_local = _beg;
                } 
             }

           } else {
             
             _beg = nextj->get_beg();
             _end = nextj->get_end();
 
             _beg_local=_end_local=_beg;

         }

         if(_beg_local2 < _end_local2) {
           
             int i1 = nextj->get_beg2() ;
             int i2 = nextj->get_end2() ;
             int treated_elts_size = _beg_local2-i2;
             int untreated_elts_size = _end_local2 - _beg_local2;

           int k= nextj->get_end();//AJOUT
           bool end_treatment = ((k==_init_end) || (i1==_init_end));//AJOUT
           if( ( (i1 < i2) && (treated_elts_size > 0)) || end_treatment) { //AJOUT

             #ifdef PRINT2
             ++nb_treatment_right; //AJOUT
             #endif

             int n_swap; int i; int tmp;
        
             if(untreated_elts_size < treated_elts_size)  {
                        n_swap = untreated_elts_size; i=_beg_local2; tmp=i2+n_swap;
             }
             else { n_swap = treated_elts_size; i=tmp=_end_local2-treated_elts_size;}

             std::swap_ranges(_first+i2, _first+i2+n_swap, _first+i);
             _beg2 = i1; _end2 = tmp;
             _beg_local2 = _end_local2 = _end2;
            #ifdef PRINT2
             nb_swap_total +=n_swap; //AJOUT
            #endif
           } else {
               if(treated_elts_size == 0) {
                _beg2 = i1; _end2 = _end_local2;
                _beg_local2 = _end_local2 = _end2;
               }
           }
           
         } else {

          _beg2 = nextj->get_beg2();
          _end2 = nextj->get_end2();
          _beg_local2=_end_local2=_end2;


        }


        if((_beg2==_init_end) || (_end==_init_end)) {
         
          #ifdef PRINT2 
          static int compteur=0; 
          std::cout <<  ++compteur <<"----------------------Remaining work -----------------------" << std::endl;
          std::cout << "nombre de sauts =" << nb_sauts << std::endl;  
          std::cout << "nb_treatment_left = " << nb_treatment_left << std::endl;
          std::cout << "nb_treatment_right = " << nb_treatment_right << std::endl;
          std::cout << "nb_swap_total = " << nb_swap_total << std::endl;
          std::cout << "before : left seq remaining [beg, end[= " << "[" << _beg << "," <<  _end << "[" << std::endl;
          std::cout << "before : right seq remaining [beg2, end2[= " << "[" << _beg2 << "," <<  _end2 << "[" << std::endl;
          #endif

          if((_end_local2-_beg_local) > _pargrain) {
               _beg=_beg_local;
               _end = _beg2 = (_end_local2 - (_end_local2-_beg_local)/2);
               _end2 = _end_local2;
              
              _end_local=_beg;
              _beg_local2=_end_local2=_end2;
              _init_end = _end;
          }
          else if((_end_local2-_beg_local) > _pargrain) {
                  _beg = _beg_local;
                  _end = _beg2 = (_end_local2 - (_end_local2-_beg_local)/2);
          
                  _end_local=_beg;
                  _beg_local2=_end_local2=_end2;
                  _init_end =  _end;

           } else {
              _beg = _beg_local;
              _end = _end_local2;

              _beg_local2=_end_local2;
              _beg2=_end2=_end_local2;
               //cout << "--------------Fin du calcul------------" << endl; 
           }
           /*cout << "after : left seq remaining [beg, end[= " << "[" << _beg << "," <<  _end << "[" << endl;
           cout << "after : right seq remaining [beg2, end2[= " << "[" << _beg2 << "," <<  _end2 << "[" << endl;
           cout << "---------------------------------------------------------------"  << endl;*/
        }

     }


};

///////////////////////////////Parallel adaptive partition with fixed grain size//////////////////
size_t block_size=3000;
typedef enum { LEFT, RIGHT, BOTH} SIDE;

template<class RandomAccessIterator>
class localwork
{
    RandomAccessIterator _data;
    bool _is_treated;
    bool EMPTY;
    size_t _s, _e;

    public :

    localwork():_data(NULL) { _is_treated=false; EMPTY=true; _s=0; _e=0 ; }
    localwork(RandomAccessIterator& data):_data(data), _is_treated(false) {EMPTY=true; _s=0; _e=0 ;}
    bool istreated() { return _is_treated;}
    void treated(bool is_treated){  _is_treated = is_treated;}
    size_t getStart() { return _s;}
    size_t getEnd() { return _e;}
    void setStart(size_t s) { _s = s;}
    void setEnd(size_t e) { _e = e;}

   template <class T1, class Predicate> friend class shared_array;
};

template <class RandomAccessIterator>
class remainingBlocks {
   
   std::vector< localwork<RandomAccessIterator> > remainingBlocksLeft;
   std::vector< localwork<RandomAccessIterator>  > remainingBlocksRight;

   public:
    remainingBlocks() { }

  template <typename RandomAccessIterator_, typename Predicate> friend class shared_array;

};

template <typename RandomAccessIterator, typename Predicate>
class shared_array {

    RandomAccessIterator  _first;
    RandomAccessIterator  _last;
    Predicate _pred;
    size_t *LN;
    size_t *RN;
    size_t *array_left;
    size_t *array_right;
    bool *isWellPlacedLeft;
    bool *isWellPlacedRight;
    remainingBlocks<RandomAccessIterator> *rb;
    pthread_mutex_t *mutex;

  public:
    shared_array()
    {
    }

    shared_array(RandomAccessIterator first, RandomAccessIterator last, Predicate pred, 
           size_t *init1, size_t* init2, size_t *start, size_t *end): _first(first), _last(last), _pred(pred)
    {
        LN = init1;
        RN = init2;
        array_left= start;
        array_right = end;
        rb = new remainingBlocks<RandomAccessIterator>();
        mutex = new pthread_mutex_t();
        pthread_mutex_init(mutex, NULL);
    }

    ~shared_array()
    {
    }

   void GetBlockFromLeftEnd(localwork<RandomAccessIterator>& leftblock)
    {
       pthread_mutex_lock(mutex);
        if((*(array_left) + block_size) <= *array_right){
           leftblock._data = _first+*array_left;
           leftblock.setStart(*array_left);
           *(array_left) +=block_size;
           leftblock.setEnd(*array_left);
           leftblock.EMPTY=false;
        } else leftblock.EMPTY=true;
       pthread_mutex_unlock(mutex);
    }

    void GetBlockFromRightEnd(localwork<RandomAccessIterator>& rightblock)
    {
       pthread_mutex_lock(mutex);
           if((*(array_right)-block_size) >= *array_left){
              *(array_right)-=block_size;
              rightblock._data = _first+*array_right;
              rightblock.setStart(*array_right);
              rightblock.setEnd(*array_right+block_size);
              rightblock.EMPTY=false;
           } else rightblock.EMPTY=true;
       pthread_mutex_unlock(mutex);
    }

    bool is_empty() {
         pthread_mutex_lock(mutex);
          bool is_empty = ((*(array_left) + block_size) > *array_right);
         pthread_mutex_unlock(mutex);
        return is_empty;
    }

   SIDE treat(localwork<RandomAccessIterator>& leftblock, localwork<RandomAccessIterator>& rightblock)
   {
       size_t left_end = leftblock.getEnd();
       size_t right_end = rightblock.getEnd();
       RandomAccessIterator  left_border = _first+left_end;
       RandomAccessIterator  right_border = _first+right_end;
       do {
            while((leftblock._data!=left_border)&&_pred(*leftblock._data)) leftblock._data++;
            while((rightblock._data!=right_border)&&(!_pred(*rightblock._data))) rightblock._data++;
            if((leftblock._data==left_border) || (rightblock._data==right_border)) break;
            std::swap(*leftblock._data, *rightblock._data);
            leftblock._data++; rightblock._data++;
       } while((leftblock._data!=left_border)&&(rightblock._data!=right_border));
       if((leftblock._data==left_border) && (rightblock._data==right_border))
               return BOTH;
       if(leftblock._data==left_border)
               return LEFT;
     return RIGHT;
    }

   void localcompute() {

         localwork<RandomAccessIterator> leftblock;
         GetBlockFromLeftEnd(leftblock);
         localwork<RandomAccessIterator> rightblock;
         GetBlockFromRightEnd(rightblock);
         size_t leftcounter = 0;
         size_t rigtcounter = 0;
         SIDE side;
         size_t s1 = leftblock.getStart() ;
         size_t e1 = leftblock.getEnd();
         size_t s2 = rightblock.getStart();
         size_t e2 = rightblock.getEnd();

         if((e1-s1 > 0) && (e2-s2 > 0))
         {
           do {
                 side =  treat(leftblock, rightblock);

                 if((side==LEFT) || (side==BOTH))
                 {
                    GetBlockFromLeftEnd(leftblock);
                    leftcounter++;
                 }

                 if((side==RIGHT) || (side==BOTH))
                 {
                    GetBlockFromRightEnd(rightblock);
                    rigtcounter++;
                 }

              } while((!leftblock.EMPTY) && (!rightblock.EMPTY));
           }

           pthread_mutex_lock(mutex);
            if(!leftblock.EMPTY)
                    rb->remainingBlocksLeft.push_back(leftblock);
            else if(!rightblock.EMPTY)
                    rb->remainingBlocksRight.push_back(rightblock);
            *LN += (leftcounter*block_size);
            *RN += (rigtcounter*block_size);
           pthread_mutex_unlock(mutex);
    }

    RandomAccessIterator compute_remaining()
    {

       size_t sz = _last-_first;
       size_t cpt1=rb->remainingBlocksLeft.size();
       size_t cpt2=rb->remainingBlocksRight.size();
       size_t  l= 0;
       size_t  r = 0;
       SIDE side;

     while((l < cpt1)&&(r < cpt2))
     {
         side=treat(rb->remainingBlocksLeft.back(), rb->remainingBlocksRight.back());
         if((side==LEFT) || (side==BOTH))
         {
           *LN+=block_size;
           rb->remainingBlocksLeft.pop_back();
           l++;
         }

         if((side==RIGHT) || (side==BOTH))
         {
             *RN+=block_size;
            rb->remainingBlocksRight.pop_back();
            r++;
         }
     }

     cpt1=rb->remainingBlocksLeft.size();
     cpt2=rb->remainingBlocksRight.size();
     size_t swapstart1=0;
     size_t swapstart2=0;
     isWellPlacedLeft = new bool[cpt1];
     isWellPlacedRight = new bool[cpt2];
     for(size_t i=0; i < cpt1; i++) isWellPlacedLeft[i]=false;
     for(size_t i=0; i < cpt2; i++) isWellPlacedRight[i]=false;

     for(size_t i=0; i < cpt1; i++) {
        if(rb->remainingBlocksLeft[i].getStart() >= *LN) {
           size_t pos = rb->remainingBlocksLeft[i].getStart();
           isWellPlacedLeft[(pos-*LN)/block_size]=true;
        }

     }

    for(size_t i=0; i < cpt2; i++) {
        if(rb->remainingBlocksRight[i].getStart() < sz-*RN) {
           size_t pos=rb->remainingBlocksRight[i].getEnd();
           isWellPlacedRight[((sz-*RN-pos)/block_size)]=true;
        }
    }

    for(size_t i=0; i < cpt1; i++) {
       if(rb->remainingBlocksLeft[i].getStart() < *LN) {
          for(size_t j = 0; j < cpt1; j++) {
             if(!isWellPlacedLeft[j]){
                isWellPlacedLeft[j]=true;
                swapstart1 = *(LN)+j*block_size;
                break;
             }
          }
          size_t left_end = rb->remainingBlocksLeft[i].getEnd();
          RandomAccessIterator left_border = _first+left_end;
          std::swap_ranges(rb->remainingBlocksLeft[i]._data,
                        left_border,
                        _first+swapstart1);
       }
    }


    for(size_t i=0; i < cpt2; i++) {
        if(rb->remainingBlocksRight[i].getStart() >= sz-*RN) {
           for(size_t j = 0; j < cpt2; j++) {
               if(!isWellPlacedRight[j]){
                  isWellPlacedRight[j]=true;
                  swapstart2 = (sz-*RN)-(j+1)*block_size;
                  break;
               }
           }
          size_t right_end = rb->remainingBlocksRight[i].getEnd();
          RandomAccessIterator right_border = _first+right_end;
          std::swap_ranges(rb->remainingBlocksRight[i]._data,
                           right_border,
                           _first + swapstart2);
        }
    }

    for(int i=cpt1-1; i >= 0; i--) rb->remainingBlocksLeft.pop_back();
    for(int i=cpt2-1; i >= 0; i--) rb->remainingBlocksRight.pop_back();

    return std::partition(_first+*LN, _first+sz-*RN, _pred);
  }
};

template<typename RandomAccessIterator, typename Predicate>
class partition_task: public tbb::task
{
  shared_array<RandomAccessIterator, Predicate> sa;

  public:

  partition_task(shared_array<RandomAccessIterator, Predicate>& _sa): sa(_sa) { }

  tbb::task* execute() {
    if(!sa.is_empty()) {
      partition_task<RandomAccessIterator, Predicate> & thief1 = *new(allocate_child() ) partition_task<RandomAccessIterator, Predicate>(sa);
      set_ref_count(2);
      spawn(thief1);
      sa.localcompute();
      tbb::task::wait_for_all(); //Wait for reference count to become one, and set reference count to zero
    }
   return NULL;
  }

};

//////////////////////////////////Partition Algorithms//////////////////////////////////////////////

typedef enum { FIXE, DYNAMIC} BLOC_SELECTION;
BLOC_SELECTION PARTITION_BLOC_SELECTION = FIXE;

template<typename RandomAccessIterator, typename Predicate>
inline RandomAccessIterator _partition_with_fixe_bloc(RandomAccessIterator first, RandomAccessIterator last, 
      Predicate pred)
{

     size_t sz = last-first;
     size_t init1=0, init2=0, start = 0, end=sz;
     shared_array<RandomAccessIterator, Predicate> sa(first, last, pred, &init1, &init2, &start, &end);
   if(sz >= 2*block_size) {
      partition_task<RandomAccessIterator, Predicate> & task = *new(tbb::task::allocate_root()) partition_task<RandomAccessIterator, Predicate>(sa);
      tbb::task::spawn_root_and_wait(task);
   }
 return sa.compute_remaining();
}


template <typename RandomAccessIterator, typename Pred>
inline RandomAccessIterator _partition_with_dynamic_bloc(RandomAccessIterator first, RandomAccessIterator last, Pred p,
      int seqgrain, int pargrain) 
{
    int size = last-first;    

     PartitionWork<RandomAccessIterator, Pred>* work ;
     work = new  PartitionWork<RandomAccessIterator, Pred>(first, last, 0, size/2, size/2, p, seqgrain, pargrain);
     tbb_adapt::run_adapt(work);
  return  work->get_split();
}

//////////////////////////////////PARTITION SELECTED Algorithm////////////////////////////////
template<typename RandomAccessIterator, typename Predicate>
inline RandomAccessIterator partition(RandomAccessIterator first, RandomAccessIterator last,
      Predicate pred, int seqgrain=100, int pargrain=2)
{
  if(PARTITION_BLOC_SELECTION==FIXE) {
   return    _partition_with_fixe_bloc(first, last, pred);
  } else {
    return _partition_with_dynamic_bloc(first, last, pred, seqgrain, pargrain);   
  }
}

}
#endif
