
#include <functional>
#include <algorithm>
#include <vector>
#include <iostream>
#include "../partition/partition.h"


/*
 *
 * quicksort
 *
 */


namespace tbb_adapt {

int sort_grain_size = 10000;

template <typename T, typename cmp=std::less<T> >
struct less_than_median
{
  T _value;
  less_than_median(){};
  less_than_median(T value):_value(value){ }
  bool operator()(T value) {
     return cmp()(value, _value);
  }
};


template<class RandomAcessIterator, class Compare>
class Intro_Sort: public tbb::task {
    typedef typename std::iterator_traits<RandomAcessIterator>::value_type val_t;
    RandomAcessIterator first;
    RandomAcessIterator last;
    int depth_limit;
    Compare cmp;
    int seqgrain;
    int pargrain;
 
   public: 
    Intro_Sort(RandomAcessIterator _first, RandomAcessIterator _last , int _depth_limit, Compare _cmp, int _seqgrain, 
       int _pargrain): first(_first), last(_last), depth_limit(_depth_limit), cmp(_cmp), seqgrain(_seqgrain), 
       pargrain(_pargrain)
     { }
                
    tbb::task* execute() {
      bool no_spawn = true;
      bool wait_for_all = false;
      while( (last-first) > tbb_adapt::sort_grain_size) { 
        if(depth_limit==0) {
           std::partial_sort(first, last, last, cmp);
           return NULL;
        }
        depth_limit = depth_limit-1;
        int sz = last-first;
        val_t median = val_t(std:: __median(*first, *(first+sz/2), *(first+sz-1)));
        RandomAcessIterator split = tbb_adapt::partition(first, last, less_than_median<val_t, Compare>(median),
        seqgrain, pargrain);
        if(no_spawn) {
          Intro_Sort<RandomAcessIterator, Compare>& thief1 = *new(allocate_child() ) Intro_Sort<RandomAcessIterator, Compare>(split, last, depth_limit, cmp, seqgrain, pargrain);
          set_ref_count(2);
          spawn(thief1);
          no_spawn = false;
         } else {
         Intro_Sort<RandomAcessIterator, Compare>& thief1 = *new(allocate_additional_child_of(*this))Intro_Sort<RandomAcessIterator, Compare>(split, last, depth_limit, cmp, seqgrain, pargrain); 
         spawn(thief1);
         }
        last=split;
        wait_for_all = true;
     }
     std::sort(first, last, cmp);
     if(wait_for_all) tbb::task::wait_for_all(); 
    return NULL;
   }
};


template<class RandomAcessIterator>
inline
void sort(RandomAcessIterator first, RandomAcessIterator last, int seqgrain=100, int pargrain=2) {
   int depth_limit = 2*tbb_adapt::log_2(last-first);
   typedef typename std::iterator_traits<RandomAcessIterator>::value_type val_t;
   Intro_Sort<RandomAcessIterator, std::less<val_t> >& task  = *new(tbb::task::allocate_root()) Intro_Sort<RandomAcessIterator, std::less<val_t> >(first, last, depth_limit, std::less<val_t>(), seqgrain, pargrain);
   tbb::task::spawn_root_and_wait(task);
}

template<class RandomAcessIterator, class Compare>
inline
void sort(RandomAcessIterator first, RandomAcessIterator last, Compare cmp, int seqgrain=100, int pargrain=2) {
   int depth_limit = 2*tbb_adapt::log_2(last-first);
   typedef typename std::iterator_traits<RandomAcessIterator>::value_type val_t;
   Intro_Sort<RandomAcessIterator, Compare>& task  = *new(tbb::task::allocate_root()) Intro_Sort<RandomAcessIterator, Compare>(first, last, depth_limit, cmp, seqgrain, pargrain);
   tbb::task::spawn_root_and_wait(task);
}

}
