#ifndef _STABLE_SORT_H_
#define _STABLE_SORT_H_


#include <iostream>

#include "tbb/task_scheduler_init.h"
#include "tbb/tick_count.h"
#include "../merge/merge.h"

namespace tbb_adapt {

int seuil = 10000;

template<class In, class Compare, class Allocator>
class MERGE_ADAPT_TASK : public tbb::task{

   typedef typename Allocator::pointer prt_t;
   In A1;
   In A2;
   In B1;
   In B2;
   prt_t input;
   Compare cmp;
   int c;

  public:

   MERGE_ADAPT_TASK(In _A1, In _A2, In _B1, In _B2,  prt_t _input, Compare _cmp, int _c): A1(_A1), 
           A2(_A2), B1(_B1), B2(_B2), input(_input), cmp(_cmp), c(_c){}

   tbb::task* execute() {
     tbb_adapt::merge(A1 , A2, B1, B2, input, cmp, c);
    return NULL;
   }
};

template<class In, class Compare, class Allocator>
class MERGE_ADAPT_TASK2 : public tbb::task{

   typedef typename Allocator::pointer prt_t;
   prt_t A1;
   prt_t A2;
   prt_t B1;
   prt_t B2;
   In input;
   Compare cmp;
   int c;

  public:

   MERGE_ADAPT_TASK2(prt_t _A1, prt_t _A2, prt_t _B1, prt_t _B2,  In _input, Compare _cmp, int _c): A1(_A1), 
         A2(_A2), B1(_B1), B2(_B2), input(_input), cmp(_cmp), c(_c){}

   tbb::task* execute() {
     tbb_adapt::merge(A1 , A2, B1, B2, input, cmp, c);
    return NULL;
   }
};


template<class In, class Compare, class Allocator>
class PMERGE_SORT : public tbb::task{
 
   typedef typename Allocator::pointer prt_t;
   In input;
   prt_t tmp;
   int sz;
   Compare cmp;
   int c;
   
  public:

   PMERGE_SORT(In _input, prt_t _tmp, int _sz, Compare _cmp, int _c): input(_input), 
       tmp(_tmp), sz(_sz), cmp(_cmp), c(_c){}


 tbb::task* execute() {
  if(sz <= seuil) std::stable_sort(input, input+sz, cmp);
  else {
   int quarter = (sz/4);
   In  A, B, C, D;
   prt_t tmpA, tmpB, tmpC, tmpD;

   A = input;
   tmpA = tmp;
   B = A + quarter;
   tmpB = tmpA + quarter;
   C = B + quarter;
   tmpC = tmpB + quarter;
   D = C + quarter;
   tmpD = tmpC + quarter;

   PMERGE_SORT<In, Compare, Allocator>& thief1 = *new(allocate_child()) PMERGE_SORT<In, Compare, Allocator>
      (A, tmpA, quarter, cmp, c);
   PMERGE_SORT<In, Compare, Allocator>& thief2 = *new(allocate_child()) PMERGE_SORT<In, Compare, Allocator>
      (B, tmpB, quarter, cmp, c);   
   PMERGE_SORT<In, Compare, Allocator>& thief3 = *new(allocate_child()) PMERGE_SORT<In, Compare, Allocator>
      (C, tmpC, quarter, cmp, c);
   PMERGE_SORT<In, Compare, Allocator>& thief4 = *new(allocate_child()) PMERGE_SORT<In, Compare, Allocator>
      (D, tmpD, sz-3*quarter, cmp,c);
   set_ref_count(5);
   spawn(thief1);
   spawn(thief2);
   spawn(thief3); 
   spawn_and_wait_for_all(thief4);
 
   MERGE_ADAPT_TASK<In, Compare, Allocator>& thief5 = *new(allocate_child()) MERGE_ADAPT_TASK<In, Compare,
        Allocator>(A, A+quarter, B, B+quarter, tmpA, cmp, c);
   MERGE_ADAPT_TASK<In, Compare, Allocator>& thief6 = *new(allocate_child()) MERGE_ADAPT_TASK<In, Compare, 
        Allocator>(C, C+quarter, D, input+sz, tmpC, cmp, c);
   set_ref_count(3);
   spawn(thief5);
   spawn_and_wait_for_all(thief6);

   MERGE_ADAPT_TASK2<In, Compare, Allocator>& thief7 = *new(allocate_child()) MERGE_ADAPT_TASK2<In, Compare, 
        Allocator>(tmpA, tmpC, tmpC, tmpA+sz, A, cmp, c);
   set_ref_count(2);
   spawn_and_wait_for_all(thief7);
  }
  
  return NULL;
 }
};

///////////////////Adptive stable_sort///////////////////////////////////////////
template<typename RandomAccessIterator, typename Compare, typename Allocator>
inline void _stable_sort_(RandomAccessIterator input, 
  typename Allocator::pointer tmp, size_t size, Compare cmp, int c)
{
  PMERGE_SORT<RandomAccessIterator, Compare, Allocator>& sort_task = *new(tbb::task::allocate_root())
       PMERGE_SORT<RandomAccessIterator, Compare, Allocator>(input, tmp, size, cmp, c);
  tbb::task::spawn_root_and_wait(sort_task);
}

///////////////////stable_sort with given comparison and given allocator/////////////////////
template<typename RandomAccessIterator, typename Compare, typename Allocator>
inline void stable_sort(RandomAccessIterator first, RandomAccessIterator last, Compare cmp, int c=100,
 Allocator Alloc=std::allocator<typename std::iterator_traits<RandomAccessIterator>::value_type>())
{
  int size = last-first;
  typedef typename Allocator::pointer prt_t;
  Allocator A;
  prt_t tmp = Alloc.allocate(size);
  tbb_adapt::_stable_sort_<RandomAccessIterator, Compare, Allocator>(first, tmp, size, cmp, c);
  A.deallocate(tmp, size);
}

///////////////////stable_sort with defaul_t comparison and given allocator/////////////////////
template<typename RandomAccessIterator, typename Allocator>
inline void stable_sort(RandomAccessIterator first, RandomAccessIterator last, int c=100,
 Allocator Alloc=std::allocator<typename std::iterator_traits<RandomAccessIterator>::value_type>())
{
  typedef typename std::iterator_traits<RandomAccessIterator>::value_type val_t;
  tbb_adapt::stable_sort<RandomAccessIterator, std::less<val_t>, Allocator>(first, last, std::less<val_t>(), c);
}

////////////////////stable_sort with given comparison and default allocator/////////////////////
template<typename RandomAccessIterator, typename Compare>
inline void stable_sort(RandomAccessIterator first, RandomAccessIterator last, Compare cmp, int c=100)
{
  typedef typename std::iterator_traits<RandomAccessIterator>::value_type val_t;
  typename  std::allocator<val_t> default_alloc;
  tbb_adapt::stable_sort<RandomAccessIterator, Compare, std::allocator<val_t> >(first, last,
      cmp, c, default_alloc);
  
}

////////////////////stable_sort with defaul comparison and default allocator/////////////////////
template<typename RandomAccessIterator>
inline void stable_sort(RandomAccessIterator first, RandomAccessIterator last, int c=100)
{
  typedef typename std::iterator_traits<RandomAccessIterator>::value_type val_t;
  typename  std::allocator<val_t> default_alloc;
  tbb_adapt::stable_sort<RandomAccessIterator, std::less<val_t>, std::allocator<val_t> >(first, last, 
      std::less<val_t>(), c, default_alloc);
}

}
#endif
